@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2016 David R. MacIver
+# Most of this work is copyright (C) 2013-2017 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -20,8 +20,8 @@ from __future__ import division, print_function, absolute_import
 from enum import IntEnum
 
 from hypothesis.errors import Frozen, InvalidArgument
-from hypothesis.internal.compat import hbytes, text_type, int_to_bytes, \
-    benchmark_time, unicode_safe_repr, reasonable_byte_type
+from hypothesis.internal.compat import hbytes, hrange, text_type, \
+    int_to_bytes, benchmark_time, unicode_safe_repr
 
 
 def uniform(random, n):
@@ -41,17 +41,18 @@ class StopTest(BaseException):
         super(StopTest, self).__init__(repr(testcounter))
         self.testcounter = testcounter
 
+
 global_test_counter = 0
 
 
-class TestData(object):
+class ConjectureData(object):
 
     @classmethod
     def for_buffer(self, buffer):
-        return TestData(
+        return ConjectureData(
             max_length=len(buffer),
             draw_bytes=lambda data, n, distribution:
-            buffer[data.index:data.index + n]
+            hbytes(buffer[data.index:data.index + n])
         )
 
     def __init__(self, max_length, draw_bytes):
@@ -74,11 +75,12 @@ class TestData(object):
         global_test_counter += 1
         self.start_time = benchmark_time()
         self.events = set()
+        self.bind_points = set()
 
     def __assert_not_frozen(self, name):
         if self.frozen:
             raise Frozen(
-                'Cannot call %s on frozen TestData' % (
+                'Cannot call %s on frozen ConjectureData' % (
                     name,))
 
     @property
@@ -104,6 +106,19 @@ class TestData(object):
             if not self.frozen:
                 self.stop_example()
 
+    def mark_bind(self):
+        """Marks a point as somewhere that a bind occurs - that is, data
+        drawn after this point may depend significantly on data drawn prior
+        to this point.
+
+        Having points like this explicitly marked allows for better shrinking,
+        as we run a pass which tries to shrink the byte stream prior to a bind
+        point while rearranging what comes after somewhat to allow for more
+        flexibility. Trying that at every point in the data stream is
+        prohibitively expensive, but trying it at a couple dozen is basically
+        fine."""
+        self.bind_points.add(self.index)
+
     def start_example(self):
         self.__assert_not_frozen('start_example')
         self.interval_stack.append(self.index)
@@ -132,7 +147,7 @@ class TestData(object):
         self.finish_time = benchmark_time()
         # Intervals are sorted as longest first, then by interval start.
         for l in self.intervals_by_level:
-            for i in range(len(l) - 1):
+            for i in hrange(len(l) - 1):
                 if l[i][1] == l[i + 1][0]:
                     self.intervals.append((l[i][0], l[i + 1][1]))
         self.intervals = sorted(
@@ -143,7 +158,22 @@ class TestData(object):
         self.events = frozenset(self.events)
         del self._draw_bytes
 
-    def draw_bytes(self, n, distribution=uniform):
+    def draw_bytes(self, n):
+        return self.draw_bytes_internal(n, uniform)
+
+    def write(self, string):
+        assert isinstance(string, hbytes)
+
+        def distribution(random, n):
+            assert n == len(string)
+            return string
+
+        while True:
+            x = self.draw_bytes_internal(len(string), distribution)
+            if x == string:
+                return
+
+    def draw_bytes_internal(self, n, distribution):
         if n == 0:
             return hbytes(b'')
         self.__assert_not_frozen('draw_bytes')
@@ -160,7 +190,7 @@ class TestData(object):
         assert self.index == initial
         self.buffer.extend(result)
         self.intervals.append((initial, self.index))
-        return reasonable_byte_type(result)
+        return hbytes(result)
 
     def mark_interesting(self):
         self.__assert_not_frozen('mark_interesting')
