@@ -3,7 +3,7 @@
 # This file is part of Hypothesis, which may be found at
 # https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2016 David R. MacIver
+# Most of this work is copyright (C) 2013-2017 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -18,15 +18,17 @@
 from __future__ import division, print_function, absolute_import
 
 import math
-from decimal import Decimal
+import datetime as dt
+import operator
+from decimal import Decimal, InvalidOperation
 from numbers import Rational
 from fractions import Fraction
 
 from hypothesis.errors import InvalidArgument
 from hypothesis.control import assume
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import ArgSpec, text_type, getargspec, \
-    integer_types, float_to_decimal
+from hypothesis.internal.compat import hrange, text_type, integer_types, \
+    getfullargspec, implements_iterator
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
 from hypothesis.utils.conventions import not_set
@@ -40,14 +42,15 @@ __all__ = [
     'choices', 'streaming',
     'booleans', 'integers', 'floats', 'complex_numbers', 'fractions',
     'decimals',
-    'characters', 'text', 'binary',
-    'tuples', 'lists', 'sets', 'frozensets',
+    'characters', 'text', 'binary', 'uuids',
+    'tuples', 'lists', 'sets', 'frozensets', 'iterables',
     'dictionaries', 'fixed_dictionaries',
     'sampled_from', 'permutations',
+    'datetimes', 'dates', 'times', 'timedeltas',
     'builds',
     'randoms', 'random_module',
     'recursive', 'composite',
-    'shared', 'runner',
+    'shared', 'runner', 'data',
 ]
 
 _strategies = set()
@@ -310,27 +313,18 @@ def floats(
     elif min_value is not None and max_value is not None:
         if min_value == max_value:
             return just(min_value)
-        elif math.isinf(max_value - min_value):
-            assert min_value < 0 and max_value > 0
-            return floats(min_value=0, max_value=max_value) | floats(
-                min_value=min_value, max_value=0
-            )
+        elif is_negative(min_value):
+            if is_negative(max_value):
+                return floats(min_value=-max_value, max_value=-min_value).map(
+                    operator.neg
+                )
+            else:
+                return floats(min_value=0.0, max_value=max_value) | floats(
+                    min_value=0.0, max_value=-min_value).map(operator.neg)
         elif count_between_floats(min_value, max_value) > 1000:
             return FixedBoundedFloatStrategy(
                 lower_bound=min_value, upper_bound=max_value
             )
-        elif is_negative(max_value):
-            assert is_negative(min_value)
-            ub_int = float_to_int(max_value)
-            lb_int = float_to_int(min_value)
-            assert ub_int <= lb_int
-            return integers(min_value=ub_int, max_value=lb_int).map(
-                int_to_float
-            )
-        elif is_negative(min_value):
-            return floats(min_value=min_value, max_value=-0.0) | floats(
-                min_value=0, max_value=max_value
-            )
         else:
             ub_int = float_to_int(max_value)
             lb_int = float_to_int(min_value)
@@ -342,7 +336,7 @@ def floats(
         if min_value < 0:
             result = floats(
                 min_value=0.0
-            ) | floats(min_value=min_value, max_value=0.0)
+            ) | floats(min_value=min_value, max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
@@ -358,7 +352,7 @@ def floats(
             result = floats(
                 min_value=0.0,
                 max_value=max_value,
-            ) | floats(max_value=0.0)
+            ) | floats(max_value=-0.0)
         else:
             result = (
                 floats(allow_infinity=allow_infinity, allow_nan=False).map(
@@ -412,13 +406,13 @@ def sampled_from(elements):
 
     from hypothesis.searchstrategy.misc import SampledFromStrategy, \
         JustStrategy
-    elements = tuple(iter(elements))
+    from hypothesis.internal.conjecture.utils import check_sample
+    elements = check_sample(elements)
     if not elements:
         return nothing()
     if len(elements) == 1:
         return JustStrategy(elements[0])
-    else:
-        return SampledFromStrategy(elements)
+    return SampledFromStrategy(elements)
 
 
 @cacheable
@@ -468,7 +462,8 @@ def lists(
                 'want to set unique_by)'
             ))
         else:
-            unique_by = lambda x: x
+            def unique_by(x):
+                return x
 
     if unique_by is not None:
         from hypothesis.searchstrategy.collections import UniqueListStrategy
@@ -540,6 +535,38 @@ def frozensets(elements=None, min_size=None, average_size=None, max_size=None):
     ).map(frozenset)
 
 
+@defines_strategy
+def iterables(elements=None, min_size=None, average_size=None, max_size=None,
+              unique_by=None, unique=False):
+    """This has the same behaviour as lists, but returns iterables instead.
+
+    Some iterables cannot be indexed (e.g. sets) and some do not have a
+    fixed length (e.g. generators). This strategy produces iterators,
+    which cannot be indexed and do not have a fixed length. This ensures
+    that you do not accidentally depend on sequence behaviour.
+
+    """
+    @implements_iterator
+    class PrettyIter(object):
+        def __init__(self, values):
+            self._values = values
+            self._iter = iter(self._values)
+
+        def __iter__(self):
+            return self._iter
+
+        def __next__(self):
+            return next(self._iter)
+
+        def __repr__(self):
+            return 'iter({!r})'.format(self._values)
+
+    return lists(
+        elements=elements, min_size=min_size, average_size=average_size,
+        max_size=max_size, unique_by=unique_by, unique=unique
+    ).map(PrettyIter)
+
+
 @defines_strategy
 def fixed_dictionaries(mapping):
     """Generate a dictionary of the same type as mapping with a fixed set of
@@ -552,7 +579,7 @@ def fixed_dictionaries(mapping):
 
     """
     from hypothesis.searchstrategy.collections import FixedKeysDictStrategy
-    check_type(dict, mapping)
+    check_type(dict, mapping, 'mapping')
     for v in mapping.values():
         check_strategy(v)
     for v in mapping.values():
@@ -803,27 +830,85 @@ def fractions(min_value=None, max_value=None, max_denominator=None):
 
 @cacheable
 @defines_strategy
-def decimals(min_value=None, max_value=None):
-    """Generates instances of decimals.Decimal.
+def decimals(min_value=None, max_value=None,
+             allow_nan=None, allow_infinity=None, places=None):
+    """Generates instances of decimals.Decimal, which may be:
 
-    If min_value is not None then all generated values are no less than
-    min_value.
+    - A finite rational number, between ``min_value`` and ``max_value``.
+    - Not a Number, if ``allow_nan`` is True.  None means "allow NaN, unless
+      ``min__value`` and ``max_value`` are not None".
+    - Positive or negative infinity, if ``max_value`` and ``min_value``
+      respectively are None, and ``allow_infinity`` is not False.  None means
+      "allow infinity, unless excluded by the min and max values".
 
-    If max_value is not None then all generated values are no greater than
-    max_value.
+    Note that where floats have one `NaN` value, Decimals have four: signed,
+    and either *quiet* or *signalling*.  See `the decimal module docs
+    <https://docs.python.org/3/library/decimal.html#special-values>`_ for
+    more information on special values.
+
+    If ``places`` is not None, all finite values drawn from the strategy will
+    have that number of digits after the decimal place.
 
     """
+    # Convert min_value and max_value to Decimal values, and validate args
+    check_valid_integer(places)
+    if places is not None and places < 0:
+        raise InvalidArgument('places=%r may not be negative' % places)
+
+    if min_value is not None:
+        min_value = Decimal(min_value)
+        if min_value.is_infinite() and min_value < 0:
+            if not (allow_infinity or allow_infinity is None):
+                raise InvalidArgument('allow_infinity=%r, but min_value=%r'
+                                      % (allow_infinity, min_value))
+            min_value = None
+        elif not min_value.is_finite():
+            # This could be positive infinity, quiet NaN, or signalling NaN
+            raise InvalidArgument(u'Invalid min_value=%r' % min_value)
+    if max_value is not None:
+        max_value = Decimal(max_value)
+        if max_value.is_infinite() and max_value > 0:
+            if not (allow_infinity or allow_infinity is None):
+                raise InvalidArgument('allow_infinity=%r, but max_value=%r'
+                                      % (allow_infinity, max_value))
+            max_value = None
+        elif not max_value.is_finite():
+            raise InvalidArgument(u'Invalid max_value=%r' % max_value)
     check_valid_bound(min_value, 'min_value')
     check_valid_bound(max_value, 'max_value')
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
 
-    floats_strategy = floats(min_value=min_value, max_value=max_value)
-    fractions_strategy = fractions(min_value=min_value, max_value=max_value)
+    if allow_infinity and (None not in (min_value, max_value)):
+        raise InvalidArgument('Cannot allow infinity between finite bounds')
+    # Set up a strategy for finite decimals
+    if places is not None:
+        # Fixed-point decimals are basically integers with a scale factor
 
-    return (
-        floats_strategy.map(float_to_decimal) |
-        fractions_strategy.map(lambda f: Decimal(f.numerator) / f.denominator)
-    )
+        def try_quantize(d):
+            try:
+                return d.quantize(factor)
+            except InvalidOperation:  # pragma: no cover
+                return None
+        factor = Decimal(10) ** -places
+        max_num = max_value / factor if max_value is not None else None
+        min_num = min_value / factor if min_value is not None else None
+        strat = integers(min_value=min_num, max_value=max_num)\
+            .map(lambda d: try_quantize(d * factor))\
+            .filter(lambda d: d is not None)
+    else:
+        # Otherwise, they're like fractions featuring a power of ten
+        strat = fractions(
+            min_value=min_value, max_value=max_value
+        ).map(lambda f: Decimal(f.numerator) / f.denominator)
+    # Compose with sampled_from for infinities and NaNs as appropriate
+    special = []
+    if allow_nan or (allow_nan is None and (None in (min_value, max_value))):
+        special.extend(map(Decimal, ('NaN', '-NaN', 'sNaN', '-sNaN')))
+    if allow_infinity or (allow_infinity is max_value is None):
+        special.append(Decimal('Infinity'))
+    if allow_infinity or (allow_infinity is min_value is None):
+        special.append(Decimal('-Infinity'))
+    return strat | sampled_from(special)
 
 
 def recursive(base, extend, max_leaves=100):
@@ -853,18 +938,127 @@ def recursive(base, extend, max_leaves=100):
 def permutations(values):
     """Return a strategy which returns permutations of the collection
     "values"."""
+    from hypothesis.internal.conjecture.utils import integer_range
+
     values = list(values)
     if not values:
-        return just(()).map(lambda _: [])
+        return builds(list)
 
-    def build_permutation(swaps):
-        initial = list(values)
-        for i, j in swaps:
-            initial[i], initial[j] = initial[j], initial[i]
-        return initial
-    n = len(values)
-    index = integers(0, n - 1)
-    return lists(tuples(index, index), max_size=n ** 2).map(build_permutation)
+    class PermutationStrategy(SearchStrategy):
+
+        def do_draw(self, data):
+            # Reversed Fisher-Yates shuffle. Reverse order so that it shrinks
+            # propertly: This way we prefer things that are lexicographically
+            # closer to the identity.
+            result = list(values)
+            for i in hrange(len(result)):
+                j = integer_range(data, i, len(result) - 1)
+                result[i], result[j] = result[j], result[i]
+            return result
+    return PermutationStrategy()
+
+
+@defines_strategy
+def datetimes(min_datetime=dt.datetime.min, max_datetime=dt.datetime.max,
+              timezones=none()):
+    """A strategy for generating datetimes, which may be timezone-aware.
+
+    This strategy works by drawing a naive datetime between ``min_datetime``
+    and ``max_datetime``, which must both be naive (have no timezone).
+
+    ``timezones`` must be a strategy that generates tzinfo objects (or None,
+    which is valid for naive datetimes).  A value drawn from this strategy
+    will be added to a naive datetime, and the resulting tz-aware datetime
+    returned.
+
+    .. note::
+        tz-aware datetimes from this strategy may be ambiguous or non-existent
+        due to daylight savings, leap seconds, timezone and calendar
+        adjustments, etc.  This is intentional, as malformed timestamps are a
+        common source of bugs.
+
+    :py:func:`hypothesis.extra.timezones` requires the ``pytz`` package, but
+    provides all timezones in the Olsen database.  If you also want to allow
+    naive datetimes, combine strategies like ``none() | timezones()``.
+
+    Alternatively, you can create a list of the timezones you wish to allow
+    (e.g. from the standard library, ``datetutil``, or ``pytz``) and use
+    :py:func:`sampled_from`.  Ensure that simple values such as None or UTC
+    are at the beginning of the list for proper minimisation.
+
+    """
+    # Why must bounds be naive?  In principle, we could also write a strategy
+    # that took aware bounds, but the API and validation is much harder.
+    # If you want to generate datetimes between two particular momements in
+    # time I suggest (a) just filtering out-of-bounds values; (b) if bounds
+    # are very close, draw a value and subtract it's UTC offset, handling
+    # overflows and nonexistent times; or (c) do something customised to
+    # handle datetimes in e.g. a four-microsecond span which is not
+    # representable in UTC.  Handling (d), all of the above, leads to a much
+    # more complex API for all users and a useful feature for very few.
+    from hypothesis.searchstrategy.datetime import DatetimeStrategy
+
+    check_type(dt.datetime, min_datetime, 'min_datetime')
+    check_type(dt.datetime, max_datetime, 'max_datetime')
+    if min_datetime.tzinfo is not None:
+        raise InvalidArgument('min_datetime=%r must not have tzinfo'
+                              % (min_datetime,))
+    if max_datetime.tzinfo is not None:
+        raise InvalidArgument('max_datetime=%r must not have tzinfo'
+                              % (max_datetime,))
+    check_valid_interval(min_datetime, max_datetime,
+                         'min_datetime', 'max_datetime')
+    if not isinstance(timezones, SearchStrategy):
+        raise InvalidArgument(
+            'timezones=%r must be a SearchStrategy that can provide tzinfo '
+            'for datetimes (either None or dt.tzinfo objects)' % (timezones,))
+    return DatetimeStrategy(min_datetime, max_datetime, timezones)
+
+
+@defines_strategy
+def dates(min_date=dt.date.min, max_date=dt.date.max):
+    """A strategy for dates between ``min_date`` and ``max_date``."""
+    from hypothesis.searchstrategy.datetime import DateStrategy
+
+    check_type(dt.date, min_date, 'min_date')
+    check_type(dt.date, max_date, 'max_date')
+    check_valid_interval(min_date, max_date, 'min_date', 'max_date')
+    if min_date == max_date:
+        return just(min_date)
+    return DateStrategy(min_date, max_date)
+
+
+@defines_strategy
+def times(min_time=dt.time.min, max_time=dt.time.max, timezones=none()):
+    """A strategy for times between ``min_time`` and ``max_time``.
+
+    The ``timezones`` argument is handled as for :py:func:`datetimes`.
+
+    """
+    check_type(dt.time, min_time, 'min_time')
+    check_type(dt.time, max_time, 'max_time')
+    if min_time.tzinfo is not None:
+        raise InvalidArgument('min_time=%r must not have tzinfo' % min_time)
+    if max_time.tzinfo is not None:
+        raise InvalidArgument('max_time=%r must not have tzinfo' % max_time)
+    check_valid_interval(min_time, max_time, 'min_time', 'max_time')
+    day = dt.date(2000, 1, 1)
+    return datetimes(min_datetime=dt.datetime.combine(day, min_time),
+                     max_datetime=dt.datetime.combine(day, max_time),
+                     timezones=timezones).map(lambda t: t.timetz())
+
+
+@defines_strategy
+def timedeltas(min_delta=dt.timedelta.min, max_delta=dt.timedelta.max):
+    """A strategy for timedeltas between ``min_delta`` and ``max_delta``."""
+    from hypothesis.searchstrategy.datetime import TimedeltaStrategy
+
+    check_type(dt.timedelta, min_delta, 'min_delta')
+    check_type(dt.timedelta, max_delta, 'max_delta')
+    check_valid_interval(min_delta, max_delta, 'min_delta', 'max_delta')
+    if min_delta == max_delta:
+        return just(min_delta)
+    return TimedeltaStrategy(min_delta=min_delta, max_delta=max_delta)
 
 
 @cacheable
@@ -872,13 +1066,14 @@ def composite(f):
     """Defines a strategy that is built out of potentially arbitrarily many
     other strategies.
 
-    This is intended to be used as a decorator. See the full
-    documentation for more details about how to use this function.
+    This is intended to be used as a decorator. See
+    :ref:`the full documentation for more details <composite-strategies>`
+    about how to use this function.
 
     """
 
-    from hypothesis.internal.reflection import copy_argspec
-    argspec = getargspec(f)
+    from hypothesis.internal.reflection import define_function_signature
+    argspec = getfullargspec(f)
 
     if (
         argspec.defaults is not None and
@@ -892,13 +1087,12 @@ def composite(f):
             'positional argument.'
         )
 
-    new_argspec = ArgSpec(
-        args=argspec.args[1:], varargs=argspec.varargs,
-        keywords=argspec.keywords, defaults=argspec.defaults
-    )
+    annots = {k: v for k, v in argspec.annotations.items()
+              if k in (argspec.args + argspec.kwonlyargs + ['return'])}
+    new_argspec = argspec._replace(args=argspec.args[1:], annotations=annots)
 
     @defines_strategy
-    @copy_argspec(f.__name__, new_argspec)
+    @define_function_signature(f.__name__, f.__doc__, new_argspec)
     def accept(*args, **kwargs):
         class CompositeStrategy(SearchStrategy):
 
@@ -913,6 +1107,7 @@ def composite(f):
 
                 return f(draw, *args, **kwargs)
         return CompositeStrategy()
+    accept.__module__ = f.__module__
     return accept
 
 
@@ -921,6 +1116,7 @@ def shared(base, key=None):
     base. Any two shared instances with the same key will share the same value,
     otherwise the identity of this strategy will be used. That is:
 
+    >>> s = integers()  # or any other strategy
     >>> x = shared(s)
     >>> y = shared(s)
 
@@ -943,7 +1139,7 @@ def choices():
 
     """
     from hypothesis.control import note, current_build_context
-    from hypothesis.internal.conjecture.utils import choice
+    from hypothesis.internal.conjecture.utils import choice, check_sample
 
     class Chooser(object):
 
@@ -955,7 +1151,7 @@ def choices():
         def __call__(self, values):
             if not values:
                 raise IndexError('Cannot choose from empty sequence')
-            result = choice(self.data, values)
+            result = choice(self.data, check_sample(values))
             with self.build_context.local():
                 self.choice_count += 1
                 note('Choice #%d: %r' % (self.choice_count, result))
@@ -1024,10 +1220,12 @@ def data():
     """This isn't really a normal strategy, but instead gives you an object
     which can be used to draw data interactively from other strategies.
 
-    It can only be used within @given, not find. This is because the lifetime
+    It can only be used within :func:`@given <hypothesis.given>`, not
+    :func:`find() <hypothesis.find>`. This is because the lifetime
     of the object cannot outlast the test body.
 
-    See the rest of the documentation for more complete information.
+    See :ref:`the rest of the documentation <interactive-draw>` for more
+    complete information.
 
     """
     from hypothesis.control import note
@@ -1041,11 +1239,14 @@ def data():
         def __repr__(self):
             return 'data(...)'
 
-        def draw(self, strategy):
+        def draw(self, strategy, label=None):
             self.data.mark_bind()
             result = self.data.draw(strategy)
             self.count += 1
-            note('Draw %d: %r' % (self.count, result))
+            if label is not None:
+                note('Draw %d (%s): %r' % (self.count, label, result))
+            else:
+                note('Draw %d: %r' % (self.count, result))
             return result
 
     class DataStrategy(SearchStrategy):
@@ -1081,15 +1282,17 @@ def data():
 # Private API below here
 
 
-def check_type(typ, arg):
+def check_type(typ, arg, name=''):
+    if name:
+        name += '='
     if not isinstance(arg, typ):
         if isinstance(typ, type):
             typ_string = typ.__name__
         else:
             typ_string = 'one of %s' % (
                 ', '.join(t.__name__ for t in typ))
-        raise InvalidArgument(
-            'Expected %s but got %r' % (typ_string, arg,))
+        raise InvalidArgument('Expected %s but got %s%r (type=%s)'
+                              % (typ_string, name, arg, type(arg).__name__))
 
 
 def check_strategy(arg):
